from dataclasses import dataclass
from datetime import timedelta
from typing import Generic, ParamSpec, TypeVar

from redis.asyncio import Redis

P = ParamSpec("P")
R = TypeVar("R")
WD = TypeVar("WD")


@dataclass
class StreamMessage:
    message_id: str
    task_id: str
    score: int


@dataclass
class WrappedContext(Generic[WD]):
    deps: WD
    redis: Redis
    task_id: str
    timeout: timedelta | int | None
    tries: int
    ttl: timedelta | int | None
    worker_id: str



import asyncio
import hashlib
from contextlib import AbstractAsyncContextManager
from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import Enum
from time import time
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Concatenate,
    Generic,
)
from uuid import UUID, uuid4

from crontab import CronTab
from redis import WatchError
from redis.asyncio import Redis
from redis.typing import EncodableT

from streaq import logger
from streaq.constants import (
    DEFAULT_TTL,
    REDIS_ABORT,
    REDIS_CHANNEL,
    REDIS_MESSAGE,
    REDIS_QUEUE,
    REDIS_RESULT,
    REDIS_RETRY,
    REDIS_RUNNING,
    REDIS_STREAM,
    REDIS_TASK,
)
from streaq.types import P, R, WD, WrappedContext
from streaq.utils import StreaqError, datetime_ms, now_ms, to_seconds, to_ms

if TYPE_CHECKING:
    from streaq.worker import Worker


class TaskStatus(str, Enum):
    PENDING = "pending"
    QUEUED = "queued"
    RUNNING = "running"
    SCHEDULED = "scheduled"
    DONE = "done"


@dataclass
class TaskData:
    fn_name: str
    args: tuple[Any, ...]
    kwargs: dict[str, Any]
    enqueue_time: int
    task_id: str | None = None
    task_try: int | None = None
    scheduled: datetime | None = None


@dataclass
class TaskResult(Generic[R]):
    success: bool
    result: R | Exception
    start_time: int
    finish_time: int
    queue_name: str


@dataclass
class Task(Generic[R]):
    """
    Represents a job that has been enqueued or scheduled.
    """

    args: tuple[Any, ...]
    kwargs: dict[str, Any]
    parent: "RegisteredCron | RegisteredTask"
    id: str = ""

    def __post_init__(self):
        if self.id:
            return
        if self.parent.unique:
            deterministic_hash = hashlib.sha256(
                self.parent.fn_name.encode()
            ).hexdigest()
            self.id = UUID(bytes=bytes.fromhex(deterministic_hash[:32]), version=4).hex
        else:
            self.id = uuid4().hex

    async def start(
        self, delay: timedelta | int | None = None, schedule: datetime | None = None
    ) -> "Task[R]":
        """

        :param delay: duration to wait before running the job
        :param schedule: datetime at which to run the job
        """
        if delay and schedule:
            raise StreaqError(
                "Use either 'delay' or 'schedule' when enqueuing tasks, not both!"
            )
        ttl_ms = to_ms(self.parent.ttl) if self.parent.ttl is not None else None
        async with self.redis.pipeline(transaction=True) as pipe:
            key = self.task_key(REDIS_TASK)
            await pipe.watch(key)
            if await pipe.exists(key, self.task_key(REDIS_RESULT)):
                await pipe.reset()
                logger.debug("Task is unique and already exists, not enqueuing!")
                return self
            enqueue_time = now_ms()
            if schedule:
                score = datetime_ms(schedule)
            elif delay is not None:
                score = enqueue_time + to_ms(delay)
            else:
                score = None
            ttl = ttl_ms or (score or enqueue_time) - enqueue_time + DEFAULT_TTL
            data = self.serialize(enqueue_time)
            pipe.multi()
            pipe.psetex(key, ttl, data)
            if score is not None:
                pipe.zadd(self.queue + REDIS_QUEUE, {self.id: score})
            else:
                await self.parent.worker.scripts["publish_task"](
                    keys=[
                        self.queue + REDIS_STREAM,
                        self.task_key(REDIS_MESSAGE),
                    ],
                    args=[self.id, enqueue_time, ttl],
                    client=pipe,
                )
            try:
                await pipe.execute()
            except WatchError:
                logger.debug("Task is unique and already exists, not enqueuing!")

        return self

    def __await__(self):
        return self.start().__await__()

    async def status(self) -> TaskStatus:
        async with self.redis.pipeline(transaction=True) as pipe:
            pipe.exists(self.task_key(REDIS_RESULT))
            pipe.exists(self.task_key(REDIS_RUNNING))
            pipe.zscore(self.queue + REDIS_QUEUE, self.id)
            pipe.exists(self.task_key(REDIS_MESSAGE))
            is_complete, is_in_progress, score, queued = await pipe.execute()

        if is_complete:
            return TaskStatus.DONE
        elif is_in_progress:
            return TaskStatus.RUNNING
        elif queued:
            return TaskStatus.QUEUED
        elif score:
            return TaskStatus.SCHEDULED
        return TaskStatus.PENDING

    def task_key(self, mid: str) -> str:
        return self.queue + mid + self.id

    def serialize(self, enqueue_time: int) -> EncodableT:
        try:
            return self.parent.worker.serializer(
                {
                    "f": self.parent.fn_name,
                    "a": self.args,
                    "k": self.kwargs,
                    "t": enqueue_time,
                }
            )
        except Exception as e:
            raise StreaqError(f"Unable to serialize task {self.parent.fn_name}:") from e

    async def result(self, timeout: timedelta | int | None = None) -> TaskResult[R]:
        raw = await self.redis.get(self.task_key(REDIS_RESULT))
        if raw is None:
            timeout_seconds = to_seconds(timeout) if timeout is not None else None
            await asyncio.wait_for(self._listen_for_result(), timeout_seconds)
            raw = await self.redis.get(self.task_key(REDIS_RESULT))
        try:
            data = self.parent.worker.deserializer(raw)
            return TaskResult(
                success=data["s"],
                result=data["r"],
                start_time=data["st"],
                finish_time=data["ft"],
                queue_name=self.queue,
            )
        except Exception as e:
            raise StreaqError(
                f"Unable to deserialize result for task {self.id}:"
            ) from e

    async def _listen_for_result(self) -> None:
        pubsub = self.redis.pubsub()
        await pubsub.subscribe(self.queue + REDIS_CHANNEL)
        encoded_id = self.id.encode()
        async for msg in pubsub.listen():
            if msg.get("data") == encoded_id:
                break

    async def abort(self, timeout: timedelta | int | None = None) -> bool:
        await self.redis.sadd(self.queue + REDIS_ABORT, self.id)  # type: ignore
        try:
            await self.result(timeout=timeout)
        except asyncio.CancelledError:
            return True
        else:
            return False

    async def info(self) -> TaskData:
        async with self.redis.pipeline(transaction=False) as pipe:
            pipe.get(self.task_key(REDIS_TASK))
            pipe.get(self.task_key(REDIS_RETRY))
            pipe.zscore(self.queue + REDIS_QUEUE, self.id)
            raw, task_try, score = await pipe.execute()
        data = self.parent.worker.deserializer(raw)
        dt = datetime.fromtimestamp(score, tz=self.parent.worker.tz) if score else None
        return TaskData(
            fn_name=self.parent.fn_name,
            args=self.args,
            kwargs=self.kwargs,
            enqueue_time=data["t"],
            task_id=self.id,
            task_try=task_try,
            scheduled=dt,
        )

    def depends(self):
        return self

    async def then(self): ...

    @property
    def redis(self) -> Redis:
        return self.parent.worker.redis

    @property
    def queue(self) -> str:
        return self.parent.worker.queue_name


@dataclass
class RegisteredTask(Generic[WD, P, R]):
    fn: Callable[Concatenate[WrappedContext[WD], P], Awaitable[R]]
    lifespan: Callable[[WrappedContext[WD]], AbstractAsyncContextManager[None]]
    max_tries: int | None
    timeout: timedelta | int | None
    ttl: timedelta | int | None
    unique: bool
    worker: "Worker"

    # TODO: add exception handler
    # TODO: add dependency injection
    # TODO: optimizations, __slots__?

    def enqueue(
        self,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Task[R]:
        """
        Serialize the job and send it to the queue for later execution by an active worker.
        Though this isn't async, it should be awaited as it returns an object that should be.
        """
        return Task(args, kwargs, self)

    async def run(self, *args: P.args, **kwargs: P.kwargs) -> R:
        """
        Run the task directly with the given params and return the result.
        This skips enqueuing and result storing in Redis.
        """
        deps = WrappedContext(
            deps=self.worker.deps,
            redis=self.worker.redis,
            task_id=uuid4().hex,
            timeout=self.timeout,
            tries=1,
            ttl=self.ttl,
            worker_id=self.worker.id,
        )
        async with self.lifespan(deps):
            try:
                return await asyncio.wait_for(
                    self.fn(deps, *args, **kwargs),
                    to_seconds(self.timeout) if self.timeout else None,
                )
            except asyncio.TimeoutError as e:
                raise e

    @property
    def fn_name(self) -> str:
        return self.fn.__qualname__

    def __repr__(self):
        return (
            f"<Task fn={self.fn_name} lifespan={self.lifespan} timeout={self.timeout} "
            f"ttl={self.ttl}>"
        )


@dataclass
class RegisteredCron(Generic[WD, R]):
    fn: Callable[[WrappedContext[WD]], Awaitable[R]]
    lifespan: Callable[[WrappedContext[WD]], AbstractAsyncContextManager[None]]
    max_tries: int | None
    run_at_startup: bool
    crontab: CronTab
    timeout: timedelta | int | None
    ttl: timedelta | int | None
    unique: bool
    worker: "Worker"

    def enqueue(self) -> Task[R]:
        """
        Serialize the job and send it to the queue for later execution by an active worker.
        Though this isn't async, it should be awaited as it returns an object that should be.
        """
        return Task((), {}, self)

    async def run(self) -> R:
        """
        Run the task directly and return the result.
        This skips enqueuing and result storing in Redis.
        """
        deps = WrappedContext(
            deps=self.worker.deps,
            redis=self.worker.redis,
            task_id=uuid4().hex,
            timeout=self.timeout,
            tries=1,
            ttl=self.ttl,
            worker_id=self.worker.id,
        )
        async with self.lifespan(deps):
            try:
                return await asyncio.wait_for(
                    self.fn(deps),
                    to_seconds(self.timeout) if self.timeout else None,
                )
            except asyncio.TimeoutError as e:
                raise e

    @property
    def fn_name(self) -> str:
        return self.fn.__qualname__

    def schedule(self) -> datetime:
        """
        Datetime of next run.
        """
        return datetime.fromtimestamp(self.next() / 1000, tz=self.worker.tz)

    def next(self) -> int:
        """
        Timestamp in milliseconds of next run.
        """
        return round((time() + self.delay) * 1000)

    def _upcoming(self) -> int:
        # Timestamp in ms for run after next run.
        diff = self.crontab.next(now=self.schedule())  # type: ignore
        return self.next() + round(diff * 1000)

    @property
    def delay(self) -> float:
        return self.crontab.next(now=datetime.now(self.worker.tz))  # type: ignore

    def __repr__(self):
        return (
            f"<Cron fn={self.fn_name} lifespan={self.lifespan} timeout={self.timeout} "
            f"ttl={self.ttl} schedule={self.schedule()}>"
        )



from redis.asyncio import Redis
from redis.commands.core import AsyncScript


PUBLISH_TASK_LUA = """
local stream_key = KEYS[1]
local task_message_id_key = KEYS[2]

local task_id = ARGV[1]
local score = ARGV[2]
local task_message_id_expire_ms = ARGV[3]

local message_id = redis.call('xadd', stream_key, '*', 'task_id', task_id, 'score', score)
redis.call('set', task_message_id_key, message_id, 'px', task_message_id_expire_ms)
return message_id
"""

PUBLISH_DELAYED_TASK_LUA = """
local delayed_queue_key = KEYS[1]
local stream_key = KEYS[2]
local task_message_id_key = KEYS[3]

local task_id = ARGV[1]
local task_message_id_expire_ms = ARGV[2]

local score = redis.call('zscore', delayed_queue_key, task_id)
if score == nil or score == false then
    return 0
end

local message_id = redis.call('xadd', stream_key, '*', 'task_id', task_id, 'score', score)
redis.call('set', task_message_id_key, message_id, 'px', task_message_id_expire_ms)
redis.call('zrem', delayed_queue_key, task_id)
return 1
"""

FETCH_TASK_LUA = """
local stream_key = KEYS[1]
local task_message_id_key = KEYS[2]

local message_id = redis.call('get', task_message_id_key)
if message_id == false then
    return nil
end

local task = redis.call('xrange', stream_key, message_id, message_id)
if task == nil then
    return nil
end

return task[1]
"""


def register_scripts(redis: Redis) -> dict[str, AsyncScript]:
    return {
        "publish_task": redis.register_script(PUBLISH_TASK_LUA),
        "publish_delayed_task": redis.register_script(PUBLISH_DELAYED_TASK_LUA),
        "fetch_task": redis.register_script(FETCH_TASK_LUA),
    }



import logging.config
import os
import sys
from multiprocessing import Process
from typing import Annotated, cast

from typer import Exit, Option, Typer

from streaq import VERSION
from streaq.utils import default_log_config, import_string
from streaq.worker import Worker

cli = Typer(context_settings={"help_option_names": ["-h", "--help"]})


def version_callback(value: bool):
    if value:
        print(f"asynQ v{VERSION}")
        raise Exit()


@cli.command()
def main(
    worker_path: str,
    workers: Annotated[
        int, Option("--workers", "-w", help="Number of worker processes to spin up")
    ] = 1,
    burst: Annotated[
        bool,
        Option(help="Whether to close workers when no tasks are left in the queue"),
    ] = False,
    watch: Annotated[
        bool, Option(help="Whether to auto-reload workers upon changes detected")
    ] = False,
    verbose: Annotated[
        bool, Option(help="Whether to use logging.DEBUG instead of logging.INFO")
    ] = False,
    version: Annotated[
        bool | None,
        Option(
            "--version", "-v", callback=version_callback, help="Show installed version"
        ),
    ] = None,
):
    logging.config.dictConfig(default_log_config(False))
    sys.path.append(os.getcwd())
    worker = cast(Worker, import_string(worker_path))
    if workers > 1:
        for _ in range(workers - 1):
            Process(target=worker.run_sync).start()
    worker.run_sync()


if __name__ == "__main__":
    cli()



import time
from datetime import datetime, timedelta
from importlib import import_module
from typing import Any


class StreaqError(Exception):
    pass


def import_string(dotted_path: str) -> Any:
    """
    Taken from pydantic.utils.
    """

    try:
        module_path, class_name = dotted_path.strip(" ").rsplit(".", 1)
    except ValueError as e:
        raise ImportError(f"'{dotted_path}' doesn't look like a module path") from e

    module = import_module(module_path)
    try:
        return getattr(module, class_name)
    except AttributeError as e:
        raise ImportError(
            f"Module '{module_path}' does not define a '{class_name}' attribute"
        ) from e


def to_seconds(timeout: timedelta | int) -> int:
    if isinstance(timeout, timedelta):
        return round(timeout.total_seconds())
    return timeout


def to_ms(timeout: timedelta | int) -> int:
    if isinstance(timeout, timedelta):
        return round(timeout.total_seconds() * 1000)
    return timeout * 1000


def now_ms() -> int:
    return round(time.time() * 1000)


def datetime_ms(dt: datetime) -> int:
    return round(dt.timestamp() * 1000)


def default_log_config(verbose: bool) -> dict[str, Any]:
    """
    Setup default config. for dictConfig.

    :param verbose: level: DEBUG if True, INFO if False
    :return: dict suitable for ``logging.config.dictConfig``
    """
    log_level = "DEBUG" if verbose else "INFO"
    return {
        "version": 1,
        "disable_existing_loggers": False,
        "handlers": {
            "streaq.standard": {
                "level": log_level,
                "class": "logging.StreamHandler",
                "formatter": "streaq.standard",
            }
        },
        "formatters": {
            "streaq.standard": {
                "format": "%(asctime)s: %(message)s",
                "datefmt": "%H:%M:%S",
            }
        },
        "loggers": {"streaq": {"handlers": ["streaq.standard"], "level": log_level}},
    }



import logging

VERSION = "0.1.0"
__version__ = VERSION

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# ruff: noqa: E402

from .task import TaskStatus
from .types import WrappedContext
from .utils import StreaqError
from .worker import StreaqRetry, Worker

__all__ = [
    "StreaqError",
    "StreaqRetry",
    "TaskStatus",
    "Worker",
    "WrappedContext",
]



import asyncio
from dataclasses import dataclass
from typing import AsyncIterator
from contextlib import asynccontextmanager
from zoneinfo import ZoneInfo

from httpx import AsyncClient
from streaq.types import WrappedContext
from streaq.worker import Worker


@dataclass
class Context:
    """
    Type safe way of defining the dependencies of your tasks.

    E.g. HTTP client, database connection, settings. etc.
    """

    http_client: AsyncClient


@asynccontextmanager
async def worker_lifespan() -> AsyncIterator[Context]:
    async with AsyncClient() as http_client:
        yield Context(http_client)


worker = Worker(worker_lifespan=worker_lifespan, tz=ZoneInfo("US/Eastern"))


@worker.task()
async def foo(ctx: WrappedContext[Context], url: str) -> int:
    # ctx.deps here is of type MyWorkerDeps, that's enforced by static typing
    # FunctionContext will also provide access to a redis connection, retry count,
    # even results of other jobs etc.
    r = await ctx.deps.http_client.get(url)
    return len(r.text)


@worker.task(timeout=5)
async def sleeper(ctx: WrappedContext[Context], time: int) -> None:
    await asyncio.sleep(time)
    print("slept!")


@worker.cron("* * * * mon-fri")
async def cronjob(ctx: WrappedContext[Context]) -> None:
    print("imma cron yay")


async def main() -> None:
    async with worker:
        # run the task directly, never sending it to a queue
        # result = await foo.run("https://www.google.com/")
        # these two are equivalent, param spec means the arguments are type safe
        for url in [
            "https://microsoft.com/",
            "https://apple.com/",
            "https://amazon.com/",
            "https://google.com/",
            "https://paypal.com/",
            "https://tesla.com/",
            "https://youtube.com/",
            "https://heroku.com/",
            "https://adobe.com/",
            "https://airbnb.com/",
            "https://github.com/",
            "https://tastytrade.com/",
            "https://chatgpt.com/",
            "https://pypi.org/",
            "https://compassion.org/",
            "https://cru.org/",
            "https://linkedin.com/",
        ]:
            await foo.enqueue(url)
        task = await foo.enqueue("https://heroku.com/").start(delay=5)
        print(task)
        print(await task.result(timeout=10))

        # test error handling
        task2 = await sleeper.enqueue(7)
        print(await task2.result())


if __name__ == "__main__":
    import asyncio

    asyncio.run(main())



DEFAULT_QUEUE_NAME = "streaq"
DEFAULT_TTL = 86_400_000  # 1 day in ms
REDIS_ABORT = ":aborted"
REDIS_CHANNEL = ":channel"
REDIS_GROUP = "workers"
REDIS_HEALTH = ":health"
REDIS_LOCK = ":lock"
REDIS_MESSAGE = ":task:messages:"
REDIS_RESULT = ":task:results:"
REDIS_RETRY = ":task:retry:"
REDIS_RUNNING = ":task:running:"
REDIS_QUEUE = ":queue"
REDIS_STREAM = ":stream"
REDIS_TASK = ":task:data:"
REDIS_TIMEOUT = ":timeout"



import asyncio
from functools import partial
import pickle
import signal
from collections import defaultdict
from collections.abc import Awaitable, Callable
from contextlib import AbstractAsyncContextManager, asynccontextmanager, suppress
from datetime import timedelta, timezone, tzinfo
from signal import Signals
from time import time
from typing import Any, AsyncIterator, Concatenate, Generic
from uuid import uuid4

from crontab import CronTab
from redis.asyncio import Redis
from redis.commands.core import AsyncScript
from redis.exceptions import ResponseError
from redis.typing import EncodableT

from streaq import logger
from streaq.constants import (
    DEFAULT_QUEUE_NAME,
    DEFAULT_TTL,
    REDIS_ABORT,
    REDIS_CHANNEL,
    REDIS_GROUP,
    REDIS_HEALTH,
    REDIS_LOCK,
    REDIS_MESSAGE,
    REDIS_QUEUE,
    REDIS_RESULT,
    REDIS_RETRY,
    REDIS_RUNNING,
    REDIS_STREAM,
    REDIS_TASK,
    REDIS_TIMEOUT,
)
from streaq.lua import register_scripts
from streaq.types import P, R, WD, StreamMessage, WrappedContext
from streaq.utils import StreaqError, now_ms, to_ms, to_seconds
from streaq.task import RegisteredCron, RegisteredTask


class StreaqRetry(StreaqError):
    """
    An exception you can manually raise in your tasks to make sure the task
    is retried.

    :param delay:
        amount of time to wait before retrying the job; if none, will be the
        square of the number of attempts
    """

    def __init__(self, delay: timedelta | int | None = None):
        self.delay = delay


@asynccontextmanager
async def task_lifespan(ctx: WrappedContext[WD]) -> AsyncIterator[None]:
    yield


@asynccontextmanager
async def worker_lifespan() -> AsyncIterator[None]:
    yield


class Worker(Generic[WD]):
    def __init__(
        self,
        redis_url: str = "redis://localhost:6379",
        concurrency: int = 16,
        queue_name: str = DEFAULT_QUEUE_NAME,
        group_name: str = REDIS_GROUP,
        queue_fetch_limit: int | None = None,
        task_lifespan: Callable[
            [WrappedContext[WD]], AbstractAsyncContextManager[None]
        ] = task_lifespan,
        worker_lifespan: Callable[
            [], AbstractAsyncContextManager[WD]
        ] = worker_lifespan,
        serializer: Callable[[Any], EncodableT] = pickle.dumps,
        deserializer: Callable[[EncodableT], Any] = pickle.loads,  # type: ignore
        burst: bool = False,
        tz: tzinfo = timezone.utc,
        handle_signals: bool = False,
        health_check_interval: timedelta | int = 300,
    ):
        self.redis = Redis.from_url(redis_url)
        self.concurrency = concurrency
        self.queue_name = queue_name
        self.group_name = group_name
        self.queue_fetch_limit = queue_fetch_limit or concurrency * 4
        self.bs = asyncio.BoundedSemaphore(concurrency + 1)
        self.counters = defaultdict(lambda: 0)
        self.worker_lifespan = worker_lifespan()
        self.loop = asyncio.get_event_loop()
        self.task_lifespan = task_lifespan
        self._deps: WD | None = None
        self.scripts: dict[str, AsyncScript] = {}
        self.registry: dict[str, RegisteredCron | RegisteredTask] = {}
        self.cron_jobs: dict[str, RegisteredCron] = {}
        self.cron_schedule: dict[str, int] = {}
        self.id = uuid4().hex
        self.serializer = serializer
        self.deserializer = deserializer
        self.tasks: dict[str, asyncio.Task] = {}
        self.handle_signals = handle_signals
        self.health_check_interval = health_check_interval
        self.tz: tzinfo = tz
        self.aborting_tasks: set[str] = set()
        if handle_signals:
            self._add_signal_handler(signal.SIGINT, self.handle_signal)
            self._add_signal_handler(signal.SIGTERM, self.handle_signal)

    @property
    def deps(self) -> WD:
        if self._deps is None:
            raise StreaqError("Worker did not start correctly!")
        return self._deps

    def build_context(
        self, registered_task: RegisteredCron | RegisteredTask, id: str, tries: int = 1
    ) -> WrappedContext[WD]:
        return WrappedContext(
            deps=self.deps,
            redis=self.redis,
            task_id=id,
            timeout=registered_task.timeout,
            tries=tries,
            ttl=registered_task.ttl,
            worker_id=self.id,
        )

    def cron(
        self,
        tab: str,
        max_tries: int | None = 3,
        run_at_startup: bool = False,
        timeout: timedelta | int | None = None,
        ttl: timedelta | int | None = 0,
        unique: bool = True,
    ):
        def wrapped(
            fn: Callable[[WrappedContext[WD]], Awaitable[R]],
        ) -> RegisteredCron[WD, R]:
            task = RegisteredCron(
                fn,
                task_lifespan,
                max_tries,
                run_at_startup,
                CronTab(tab),
                timeout,
                ttl,
                unique,
                self,
            )
            self.cron_jobs[task.fn_name] = task
            self.registry[task.fn_name] = task
            return task

        return wrapped

    def task(
        self,
        max_tries: int | None = 3,
        timeout: timedelta | int | None = None,
        ttl: timedelta | int | None = timedelta(minutes=5),
        unique: bool = False,
    ):
        def wrapped(
            fn: Callable[Concatenate[WrappedContext[WD], P], Awaitable[R]],
        ) -> RegisteredTask[WD, P, R]:
            task = RegisteredTask(
                fn,
                task_lifespan,
                max_tries,
                timeout,
                ttl,
                unique,
                self,
            )
            self.registry[task.fn_name] = task
            return task

        return wrapped

    def run_sync(self) -> None:
        """
        Sync function to run the worker, finally closes worker connections.
        """
        self.main_task = self.loop.create_task(self.main())
        try:
            self.loop.run_until_complete(self.main_task)
        except asyncio.CancelledError:
            pass
        finally:
            self.loop.run_until_complete(self.close())

    async def run_async(self) -> None:
        self.main_task = self.loop.create_task(self.main())
        await self.main_task

    async def main(self):
        logger.info(f"Starting worker {self.id} for {len(self)} functions...")
        # create consumer group if it doesn't exist
        with suppress(ResponseError):
            await self.redis.xgroup_create(
                name=self.queue_name + REDIS_STREAM,
                groupname=self.group_name,
                id=0,
                mkstream=True,
            )
        # set schedule for cron jobs
        ts = now_ms()
        for name, cron_job in self.cron_jobs.items():
            if cron_job.run_at_startup and name not in self.cron_schedule:
                self.cron_schedule[name] = ts
            else:
                self.cron_schedule[name] = cron_job.next()
        # run loops
        async with self:
            done, pending = await asyncio.wait(
                [
                    asyncio.ensure_future(self.listen_queue()),
                    asyncio.ensure_future(self.listen_stream()),
                    asyncio.ensure_future(self.consumer_cleanup()),
                    asyncio.ensure_future(self.health_check()),
                    asyncio.ensure_future(self.redis_health_check()),
                ],
                return_when=asyncio.FIRST_COMPLETED,
            )
            for task in pending:
                task.cancel()
            await asyncio.gather(*pending, return_exceptions=True)
            for task in done:
                task.result()
            await self.close()

    async def consumer_cleanup(self) -> None:
        while True:
            await asyncio.sleep(300)  # every 5 minutes
            consumers_info = await self.redis.xinfo_consumers(
                self.queue_name + REDIS_STREAM,
                groupname=self.group_name,
            )
            for consumer_info in consumers_info:
                if self.id == consumer_info["name"].decode():
                    continue

                idle = timedelta(milliseconds=consumer_info["idle"]).seconds
                pending = consumer_info["pending"]

                if pending == 0 and idle > DEFAULT_TTL / 1000:
                    await self.redis.xgroup_delconsumer(
                        name=self.queue_name + REDIS_STREAM,
                        groupname=self.group_name,
                        consumername=consumer_info["name"],
                    )

    async def listen_queue(self) -> None:
        set_name = self.queue_name + REDIS_ABORT
        while True:
            start_time = time()
            async with self.redis.pipeline(transaction=True) as pipe:
                pipe.zrange(
                    self.queue_name + REDIS_QUEUE,
                    start=0,
                    end=now_ms(),
                    num=self.queue_fetch_limit,
                    offset=0,
                    withscores=True,
                    byscore=True,
                )
                pipe.smembers(set_name)
                task_ids, aborted_ids = await pipe.execute()
            async with self.redis.pipeline(transaction=False) as pipe:
                for task_id, score in task_ids:
                    expire_ms = int(score - now_ms() + DEFAULT_TTL)
                    if expire_ms <= 0:
                        expire_ms = DEFAULT_TTL

                    decoded = task_id.decode()
                    await self.scripts["publish_delayed_task"](
                        keys=[
                            self.queue_name + REDIS_QUEUE,
                            self.queue_name + REDIS_STREAM,
                            self.queue_name + REDIS_MESSAGE + decoded,
                        ],
                        args=[decoded, expire_ms],
                        client=pipe,
                    )
                # Go through job_ids in the aborted tasks set and cancel those tasks.
                aborted: set[str] = set()
                for task_id_bytes in aborted_ids:
                    task_id = task_id_bytes.decode()
                    if task_id in self.tasks:
                        self.tasks[task_id].cancel()
                        aborted.add(task_id)
                if aborted:
                    self.aborting_tasks.update(aborted)
                    pipe.srem(set_name, *aborted)
                await pipe.execute()

            # cron jobs
            job_futures = set()
            ts = now_ms()
            for name, cron_job in self.cron_jobs.items():
                diff = max(0, self.cron_schedule[name] - ts)
                if diff < 1000:
                    job_futures.add(
                        cron_job.enqueue().start(delay=timedelta(milliseconds=diff))
                    )
                    self.cron_schedule[name] = cron_job._upcoming()
            if job_futures:
                await asyncio.gather(*job_futures)

            # cleanup aborted tasks
            for task_id, task in list(self.tasks.items()):
                if task.done():
                    del self.tasks[task_id]
                    # propagate error
                    task.result()

            delay = time() - start_time
            if delay < 1:
                await asyncio.sleep(delay)

    async def listen_stream(self):
        while True:
            messages: list[StreamMessage] = []
            active_jobs = self.active
            count = self.queue_fetch_limit - active_jobs
            if active_jobs < self.concurrency:
                expired = await self.get_idle_tasks(count)
                count -= len(expired)
                messages.extend(expired)
            if count > 0:
                res = await self.redis.xreadgroup(
                    groupname=self.group_name,
                    consumername=self.id,
                    streams={self.queue_name + REDIS_STREAM: ">"},
                    block=500,
                    count=count,
                )
                for _, msgs in res:
                    messages.extend(
                        [
                            StreamMessage(
                                message_id=msg_id.decode(),
                                task_id=msg[b"task_id"].decode(),
                                score=int(msg[b"score"]),
                            )
                            for msg_id, msg in msgs
                        ]
                    )
            # start new tasks
            for message in messages:
                coro = self.run_task(message.task_id, message.message_id)
                self.tasks[message.task_id] = self.loop.create_task(coro)

    async def get_idle_tasks(self, count: int) -> list[StreamMessage]:
        ids = await self.redis.zrangebyscore(
            self.queue_name + REDIS_TIMEOUT, 0, now_ms()
        )
        if not ids:
            return []
        if len(ids) > count:
            ids = ids[:count]
        async with self.redis.pipeline(transaction=True) as pipe:
            pipe.zrem(self.queue_name + REDIS_TIMEOUT, *ids)
            pipe.xclaim(
                self.queue_name + REDIS_STREAM,
                groupname=self.group_name,
                consumername=self.id,
                min_idle_time=1000,
                message_ids=ids,
            )
            _, msgs = await pipe.execute()
        return [
            StreamMessage(
                message_id=msg_id.decode(),
                task_id=msg[b"task_id"].decode(),
                score=int(msg[b"score"]),
            )
            for msg_id, msg in msgs
        ]

    async def run_task(self, task_id: str, message_id: str):
        """
        Execute the registered task, then store the result in Redis.
        """
        key = lambda mid: self.queue_name + mid + task_id
        # acquire semaphore
        async with self.bs:
            start_time = now_ms()
            async with self.redis.pipeline(transaction=True) as pipe:
                pipe.get(key(REDIS_TASK))
                pipe.incr(key(REDIS_RETRY))
                pipe.srem(self.queue_name + REDIS_ABORT, task_id)
                pipe.expire(key(REDIS_RETRY), DEFAULT_TTL // 1000)
                raw, task_try, abort, _ = await pipe.execute()

            async def handle_failure(exc: BaseException) -> None:
                self.counters["failed"] += 1
                data = {
                    "s": False,
                    "r": exc,
                    "st": start_time,
                    "ft": now_ms(),
                }
                try:
                    raw = self.serializer(data)
                    await asyncio.shield(
                        self.finish_failed_task(task_id, message_id, raw, task.ttl)
                    )
                except Exception as e:
                    raise StreaqError(
                        f"Failed to serialize result for task {task_id}!"
                    ) from e

            if not raw:
                logger.warning(f"Task {task_id} expired.")
                return await handle_failure(StreaqError("Task execution failed!"))

            try:
                data = self.deserializer(raw)
            except Exception as e:
                logger.exception(f"Failed to deserialize task {task_id}: {e}")
                return await handle_failure(e)

            fn_name = data["f"]
            task = self.registry[fn_name]

            if abort:
                logger.info(f"Task {task_id} aborted prior to start.")
                return await handle_failure(asyncio.CancelledError())

            timeout = (
                "inf"
                if task.timeout is None
                else round(time()) + to_seconds(task.timeout) + 1
            )
            await self.redis.zadd(
                self.queue_name + REDIS_TIMEOUT, {message_id: timeout}
            )

            if task.max_tries and task_try > task.max_tries:
                logger.warning(f"Max retry attempts reached for task {task_id}!")
                return await handle_failure(
                    StreaqError(f"Max retry attempts reached for task {task_id}!")
                )

            ctx = self.build_context(task, task_id)
            async with task.lifespan(ctx):
                success = True
                delay = None
                done = True
                try:
                    logger.info(f"Task {task_id} starting in worker {self.id}...")
                    result = await asyncio.wait_for(
                        task.fn(ctx, *data["a"], **data["k"]),
                        to_seconds(task.timeout) if task.timeout is not None else None,
                    )
                except StreaqRetry as e:
                    result = e
                    success = False
                    done = False
                    if e.delay is not None:
                        delay = to_seconds(e.delay)
                    else:
                        delay = task_try**2
                    logger.info(f"Retrying task {task_id} in {delay} seconds...")
                except asyncio.TimeoutError as e:
                    logger.info(f"Task {task_id} timed out!")
                    result = e
                    success = False
                    done = True
                except asyncio.CancelledError as e:
                    if task_id in self.aborting_tasks:
                        logger.info(f"Task {task_id} aborted!")
                        self.aborting_tasks.remove(task_id)
                        done = True
                    else:
                        logger.info(f"Task {task_id} cancelled, will be run again.")
                        done = False
                    result = e
                    success = False
                except Exception as e:
                    result = e
                    success = False
                    done = True
                finally:
                    data = {
                        "s": success,
                        "r": result,  # type: ignore
                        "st": start_time,
                        "ft": now_ms(),
                    }
                    try:
                        raw = self.serializer(data)
                        await asyncio.shield(
                            self.finish_task(
                                task_id,
                                message_id,
                                done,
                                raw,
                                task.ttl,
                                delay,
                            )
                        )
                    except Exception as e:
                        raise StreaqError(
                            f"Failed to serialize result for task {task_id}!"
                        ) from e

    async def finish_task(
        self,
        task_id: str,
        message_id: str,
        finish: bool,
        result: EncodableT,
        ttl: timedelta | int | None,
        incr_score: int | None,
    ) -> None:
        key = lambda mid: self.queue_name + mid + task_id
        async with self.redis.pipeline(transaction=True) as pipe:
            delete_keys = set()
            stream_key = self.queue_name + REDIS_STREAM
            delete_keys.add(key(REDIS_RUNNING))
            pipe.xack(
                stream_key,
                self.group_name,
                message_id,
            )
            pipe.xdel(stream_key, message_id)
            pipe.publish(self.queue_name + REDIS_CHANNEL, task_id)

            if finish:
                self.counters["completed"] += 1
                if result and ttl != 0:
                    pipe.set(key(REDIS_RESULT), result, ex=ttl)
                delete_keys.add(key(REDIS_RETRY))
                delete_keys.add(key(REDIS_TASK))
                delete_keys.add(key(REDIS_MESSAGE))
                pipe.zrem(self.queue_name + REDIS_TIMEOUT, message_id)
                pipe.srem(self.queue_name + REDIS_ABORT, task_id)
            elif incr_score:
                self.counters["retried"] += 1
                delete_keys.add(key(REDIS_MESSAGE))
                pipe.zadd(self.queue_name, {message_id: now_ms() + incr_score})
            else:
                self.counters["retried"] += 1
                score = now_ms()
                ttl_ms = to_ms(ttl) if ttl is not None else None
                expire = ttl_ms or score + DEFAULT_TTL
                await self.scripts["publish_task"](
                    keys=[stream_key, key(REDIS_MESSAGE)],
                    args=[task_id, score, expire],
                    client=pipe,
                )
            if delete_keys:
                pipe.delete(*delete_keys)
            await pipe.execute()
        logger.info(f"Task {task_id} finished.")

    async def finish_failed_task(
        self,
        task_id: str,
        message_id: str,
        result_data: EncodableT,
        ttl: timedelta | int | None,
    ) -> None:
        key = lambda mid: self.queue_name + mid + task_id
        stream_key = self.queue_name + REDIS_STREAM
        self.counters["failed"] += 1
        async with self.redis.pipeline(transaction=True) as pipe:
            pipe.delete(
                key(REDIS_RETRY),
                key(REDIS_RUNNING),
                key(REDIS_TASK),
                key(REDIS_MESSAGE),
            )
            pipe.publish(self.queue_name + REDIS_CHANNEL, task_id)
            pipe.srem(self.queue_name + REDIS_ABORT, task_id)
            pipe.zrem(self.queue_name + REDIS_TIMEOUT, message_id)
            pipe.xack(stream_key, self.group_name, message_id)
            pipe.xdel(stream_key, message_id)
            if result_data is not None:
                pipe.set(key(REDIS_RESULT), result_data, ex=ttl)
            await pipe.execute()

    async def redis_health_check(self):
        timeout = to_seconds(self.health_check_interval)
        await asyncio.sleep(1)
        async with self.redis.lock(
            self.queue_name + REDIS_LOCK,
            sleep=timeout + 1,
            timeout=timeout,
        ):
            while True:
                async with self.redis.pipeline(transaction=False) as pipe:
                    pipe.info(section="Server")
                    pipe.info(section="Memory")
                    pipe.info(section="Clients")
                    pipe.dbsize()
                    pipe.xlen(self.queue_name + REDIS_STREAM)
                    pipe.zcard(self.queue_name + REDIS_QUEUE)
                    (
                        info_server,
                        info_memory,
                        info_clients,
                        key_count,
                        stream_size,
                        queue_size,
                    ) = await pipe.execute()

                redis_version = info_server.get("redis_version", "?")
                mem_usage = info_memory.get("used_memory_human", "?")
                clients_connected = info_clients.get("connected_clients", "?")
                health = (
                    f"redis_version={redis_version} "
                    f"mem_usage={mem_usage} "
                    f"clients_connected={clients_connected} "
                    f"db_keys={key_count} "
                    f"queued={stream_size} "
                    f"scheduled={queue_size}"
                )
                logger.info(health)
                await self.redis.hset(  # type: ignore
                    self.queue_name + REDIS_HEALTH, "redis", health
                )
                await asyncio.sleep(timeout)

    async def health_check(self):
        while True:
            await asyncio.sleep(to_seconds(self.health_check_interval))
            health = repr(self)[1:-1]
            logger.info(health)
            await self.redis.hset(  # type: ignore
                self.queue_name + REDIS_HEALTH, self.id, health
            )

    def _add_signal_handler(self, signum: int, handler: Callable[[int], None]) -> None:
        try:
            self.loop.add_signal_handler(signum, partial(handler, signum))
        except NotImplementedError:
            logger.error("Windows does not support handling signals for an event loop!")

    def handle_signal(self, signum: int) -> None:
        signal = Signals(signum)
        logger.info(
            f"Shutdown on {signal.name}: "
            + repr(self)[1:-1].replace("running", "interrupted")
        )
        for t in self.tasks.values():
            if not t.done():
                t.cancel()
        self.main_task.cancel()

    async def queue_size(self) -> int:
        """
        Returns the number of tasks currently queued in Redis.
        """
        async with self.redis.pipeline(transaction=True) as pipe:
            pipe.xlen(self.queue_name + REDIS_STREAM)
            pipe.zcard(self.queue_name + REDIS_QUEUE)
            stream_size, queue_size = await pipe.execute()

        return stream_size + queue_size

    @property
    def active(self) -> int:
        return sum(not t.done() for t in self.tasks.values())

    async def close(self) -> None:
        if not self.handle_signals:
            self.handle_signal(signal.SIGUSR1)
        await asyncio.gather(*self.tasks.values())
        # delete health hash
        await self.redis.delete(self.queue_name + REDIS_HEALTH)
        await self.redis.close(close_connection_pool=True)

    async def __aenter__(self):
        # register lua scripts
        self.scripts.update(register_scripts(self.redis))
        # user-defined deps
        self._deps = await self.worker_lifespan.__aenter__()
        return self

    async def __aexit__(self, *exc):
        await self.worker_lifespan.__aexit__(*exc)

    def __len__(self):
        """
        Returns the number of functions (including cron jobs) registered.
        """
        return len(self.registry)

    def __repr__(self) -> str:
        return (
            f"<Worker {self.id}: completed={self.counters['completed']} failed="
            f"{self.counters['failed']} retried={self.counters['retried']} running="
            f"{self.active}>"
        )
